
/*
   Solver for "Scramble Squares" Puzzles by Viktor "AlgorithMan" Engelmann

   This code is provided under the terms of the gnu general public license
   version 3. See gpl3.txt for details.
*/

#include<iostream>
#include<iomanip>
#include<vector>
#include<fstream>
using namespace std;


class Puzzlepiece {
	private:
		bool equal(const Puzzlepiece& x)              { return up==x.up && right==x.right && down==x.down && left==x.left; }
	public:
		short up;
		short down;
		short left;
		short right;

		Puzzlepiece(short u=0, short r=0, short d=0, short l=0) { up=u; right=r; down=d; left=l; }
		void Rotate() { *this = Puzzlepiece(left, up, right, down); }
		bool FitsRightOf (const Puzzlepiece& x) const { return((x.right == -left)||(x.right == 0)); }
		bool FitsUnder (const Puzzlepiece& x)   const { return((x.down == -up)||(x.down == 0)); }
		bool FitsLeftOf (const Puzzlepiece& x)  const { return((x.left == -right)||(x.left == 0)); }
		bool FitsOver (const Puzzlepiece& x)    const { return((x.up == -down)||(x.up == 0)); }
		bool operator=(const Puzzlepiece& x)          { up=x.up; right=x.right; down=x.down; left=x.left; }
		bool operator==(Puzzlepiece x)                { for(int i = 0; i < 4; i++)
		                                                	if(equal(x))
		                                                		return true;
		                                                	else
		                                                		x.Rotate();
		                                                return false;
		                                              }
};

istream& operator>>(istream& is, Puzzlepiece& P) {
	return is >> P.up >> P.right >> P.down >> P.left;
}

int w=3, h=3, n=9, recursions = 0;
vector<vector<Puzzlepiece> > Desk;
vector<Puzzlepiece> Pieces;
vector<int> Availiable;

void Print() {
	for(int y = 1; y <= h; y++) {
		for(int x = 1; x <= w; x++)
			cout << "  +--------+ ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  |   " << setw(2) << Desk[y][x].up << "   | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  | " << setw(2) << Desk[y][x].left << "  " << setw(2) << Desk[y][x].right << " | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  |   " << setw(2) << Desk[y][x].down << "   | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  +--------+ ";
		cout << "\n\n";
	}
	cout << "\n\n"; 
}

void Puzzle(int piecesleft) {
	if(piecesleft <= 0) {
		Print();
		return;
	}
	recursions++;

	// count possibilities
	vector<int> c_possibilities(n,0);
	vector<vector<int> > p_possibilities(h+2,vector<int>(w+2,0));
	for(int y = 1; y <= h; y++) {
		for(int x = 1; x <= w; x++) {
			if(Desk[y][x].up != 0) // if occupied, then try next
				continue;
			for(int i = 0; i < n; i++) {
				if(Availiable[i] <= 0)
					continue;
				for(int j = 0; j < 4; j++) {
					if(Pieces[i].FitsRightOf(Desk[y][x-1])
					&& Pieces[i].FitsUnder(Desk[y-1][x])
					&& Pieces[i].FitsLeftOf(Desk[y][x+1])
					&& Pieces[i].FitsOver(Desk[y+1][x])) {
						c_possibilities[i]++;
						p_possibilities[y][x]++;
					}
					if(i==0) // avoid symmetric solutions
						break;
					Pieces[i].Rotate();
				}
			}
		}
	}


	// select card or position to branch over
	bool mode = 0; // 0 = branch over a card, 1 = branch over a position
	int min_possibilities = 4*w*h*n+1, pos_info_1=-1, pos_info_2=-1;
	for(int i = 0; i < n; i++) {
		if((Availiable[i] > 0) && (c_possibilities[i] < min_possibilities)) {
			min_possibilities = c_possibilities[i];
			pos_info_1 = i;
		}
	}
	for(int y = 1; y <= h; y++) {
		for(int x = 1; x <= w; x++) {
			if(Desk[y][x].up == 0 && (p_possibilities[y][x] < min_possibilities)) {
				min_possibilities = p_possibilities[y][x];
				pos_info_1 = y;
				pos_info_2 = x;
				mode = 1;
			}
		}
	}
	if(min_possibilities == 0)
		return;

	// do the recursions
	if(!mode) {
		Availiable[pos_info_1]--;
		for(int y = 1; y <= h; y++) {
			for(int x = 1; x <= w; x++) {
				if(Desk[y][x].up != 0) // if occupied, then try next
					continue;
				Desk[y][x] = Pieces[pos_info_1];
				for(int j = 0; j < 4; j++) {
					if(Desk[y][x].FitsRightOf(Desk[y][x-1])
					&& Desk[y][x].FitsUnder(Desk[y-1][x])
					&& Desk[y][x].FitsLeftOf(Desk[y][x+1])
					&& Desk[y][x].FitsOver(Desk[y+1][x])) {
						Puzzle(piecesleft-1);
					}
					if(pos_info_1==0) // avoid symmetric solutions
						break;
					Desk[y][x].Rotate();
				}
				Desk[y][x] = Puzzlepiece(0,0,0,0);
			}
		}
		Availiable[pos_info_1]++;
	} else {
		for(int i = 0; i < n; i++) {
			if(Availiable[i] <= 0)
				continue;
			Availiable[i]--;
			Desk[pos_info_1][pos_info_2] = Pieces[i];
			for(int j = 0; j < 4; j++) {
				if(Desk[pos_info_1][pos_info_2].FitsRightOf(Desk[pos_info_1][pos_info_2-1])
				&& Desk[pos_info_1][pos_info_2].FitsUnder(Desk[pos_info_1-1][pos_info_2])
				&& Desk[pos_info_1][pos_info_2].FitsLeftOf(Desk[pos_info_1][pos_info_2+1])
				&& Desk[pos_info_1][pos_info_2].FitsOver(Desk[pos_info_1+1][pos_info_2]))
					Puzzle(piecesleft-1);
				if(i==0) // avoid symmetric solutions
					break;
				Desk[pos_info_1][pos_info_2].Rotate();
			}
			Availiable[i]++;
			Desk[pos_info_1][pos_info_2] = Puzzlepiece(0,0,0,0);
		}
	}
}

int main(int argc, char** argv) {
	if(argc < 2) {
		cerr << "usage: " << argv[0] << " <puzzle file>\n";
		return 1;
	}
	ifstream fin(argv[1]);
	fin >> h >> w;
	n = 0;

	Puzzlepiece Piece;
	for(int i = w*h; i > 0; i--) {
		fin >> Piece;
		bool prior=false;
		for(int j = 0; j < n && !prior; j++) {
			if(Piece == Pieces[j]) {
				Availiable[j]++;
				prior=true;
			}
		}
		if(!prior) {
			Availiable.push_back(1);
			Pieces.push_back(Piece);
			n++;
		}
	}
	Desk = vector<vector<Puzzlepiece> >(h+2, vector<Puzzlepiece>(w+2));
	Puzzle(w*h);
	cout << "needed " << recursions << " recursions\n";
	return(0);
}
